import argparse
import json
import sys
import re
from decimal import Decimal


class PredefinedAIArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.predefinedai_args = {}
        self.predefinedai_support_args = (
            "PredefinedAIArgumentParser.int",
            "PredefinedAIArgumentParser.float",
            "PredefinedAIArgumentParser.str",
            "PredefinedAIArgumentParser.column",
            "PredefinedAIArgumentParser.columns",
            "PredefinedAIArgumentParser.bool",
        )
        self.predefinedai_inner_func = (
            "int_parse",
            "float_parse",
            "str_parse",
            "column_parse",
            "columns_parse",
            "bool_parse",
        )

    def add_argument(self, *args, **kwargs):
        if "action" in kwargs and kwargs["action"] == "help":
            super().add_argument(*args, **kwargs)
            return

        if "type" not in kwargs or not kwargs["type"][0].__name__ in self.predefinedai_inner_func:
            raise ValueError(f"type must be in {self.predefinedai_support_args}")

        if "action" in kwargs:
            raise ValueError("Cannot use action in PredefinedAIArgumentParser.")

        parse_func, local_params = kwargs["type"]
        kwargs["type"] = parse_func
        super().add_argument(*args, **kwargs)

        if not args[0].startswith("--"):
            raise ValueError("name must start with --")

        name = args[0][2:]
        default = None
        if "default" in kwargs:
            default = kwargs["default"]

        description = None
        if "help" in kwargs:
            description = kwargs["help"]
        else:
            description = ""

        required = None
        if "required" in kwargs:
            required = kwargs["required"]
        else:
            required = False

        choices = None
        if "choices" in kwargs:
            choices = kwargs["choices"]

        arg_dict = {"name": name, "default": default, "description": description, "required": required}
        if parse_func.__name__ == "int_parse":
            if default is None:
                default = local_params["max"] - local_params["min"]

            if choices is None and local_params["window"] == 0:
                raise ValueError("Please specify choices or window.")

            if choices is not None and local_params["window"] != 0:
                raise ValueError("Cannot use choices and window both.")

            arg_dict.update({"type": "int", "min": local_params["min"], "max": local_params["max"]})
            if choices is not None:
                arg_dict["choices"] = choices
            else:
                arg_dict["window"] = local_params["window"]
        elif parse_func.__name__ == "float_parse":
            if default is None:
                default = local_params["max"] - local_params["min"]

            if choices is None and local_params["window"] == 0.0:
                raise ValueError("Please specify choices or window.")

            if choices is not None and local_params["window"] != 0.0:
                raise ValueError("Cannot use choices and window both.")

            arg_dict.update({"type": "float", "min": local_params["min"], "max": local_params["max"]})
            if choices is not None:
                arg_dict["choices"] = choices
            else:
                arg_dict["window"] = local_params["window"]
        elif parse_func.__name__ == "str_parse":
            if default is None:
                default = ""

            arg_dict.update({"type": "str"})
            if choices is not None:
                arg_dict["choices"] = choices
        elif parse_func.__name__ == "column_parse":
            if default is None:
                default = ""

            arg_dict.update({"type": "column"})
        elif parse_func.__name__ == "columns_parse":
            if default is None:
                default = "[]"

            arg_dict.update(
                {
                    "type": "columns",
                    "name": name,
                    "default": default,
                    "description": description,
                    "required": required,
                }
            )
        elif parse_func.__name__ == "bool_parse":
            if default is None:
                default = True

            arg_dict.update({"type": "bool"})
        else:
            raise argparse.ArgumentTypeError(f"parse function {parse_func.__name__} cannot found.")

        self.predefinedai_args[arg_dict["name"]] = arg_dict

    @staticmethod
    def int(min: int = 0, max: int = sys.maxsize, window: int = 1):
        if min > max:
            raise argparse.ArgumentTypeError(f"min({min}) must be less than max({max})")

        def int_parse(arg):
            if isinstance(arg, str):
                arg = re.sub("[\\\"']", "", arg)

            try:
                val = int(arg)
            except ValueError:
                raise argparse.ArgumentTypeError(f"{arg} cannot cast to integer.")

            if val < min:
                raise argparse.ArgumentTypeError(f"{arg} must be greater than {min}.")
            elif val > max:
                raise argparse.ArgumentTypeError(f"{arg} must be less than {max}.")
            elif (val - min) % window > 1e-128:
                raise argparse.ArgumentTypeError(f"'({arg} - {min}) mod {window}' must be zero.")
            return val

        return int_parse, locals()

    @staticmethod
    def float(min: float = 0.0, max: float = sys.float_info.max, window: float = 1.0):
        if min > max:
            raise argparse.ArgumentTypeError(f"min({min}) must be less than max({max})")

        def float_parse(arg):
            if isinstance(arg, str):
                arg = re.sub("[\\\"']", "", arg)

            try:
                val = float(arg)
            except ValueError:
                raise argparse.ArgumentTypeError(f"{arg} cannot cast to float.")

            if val < min:
                raise argparse.ArgumentTypeError(f"{arg} must be greater than {min}.")
            elif val > max:
                raise argparse.ArgumentTypeError(f"{arg} must be less than {max}.")
            elif ((Decimal(str(val)) - Decimal(str(min))) % Decimal(str(window))) != 0:
                raise argparse.ArgumentTypeError(f"'({arg} - {min}) mod {window}' must be zero.")
            return val

        return float_parse, locals()

    @staticmethod
    def str():
        def str_parse(arg):
            try:
                val = str(arg)
            except ValueError:
                raise argparse.ArgumentTypeError(f"{arg} cannot cast to str.")

            return val

        return str_parse, locals()

    @staticmethod
    def column():
        def column_parse(arg):
            try:
                val = str(arg)
            except ValueError:
                raise argparse.ArgumentTypeError(f"{arg} cannot cast to str. column must be str.")

            return val

        return column_parse, locals()

    @staticmethod
    def columns():
        def columns_parse(arg):
            try:
                val = json.loads(arg)
            except Exception as e:
                print(e)
                raise argparse.ArgumentTypeError(
                    f"{arg} cannot cast to dictionary. columns must be json string."
                )

            return val

        return columns_parse, locals()

    @staticmethod
    def bool():
        def bool_parse(arg):
            if isinstance(arg, str):
                arg = re.sub("[\\\"']", "", arg)

            if isinstance(arg, bool):
                return arg
            if arg.lower() in ("yes", "true", "t", "y", "1"):
                return True
            elif arg.lower() in ("no", "false", "f", "n", "0"):
                return False
            else:
                raise argparse.ArgumentTypeError("Boolean value expected.")

        return bool_parse, locals()



from mlplatform_lib.mlplatform.mlplatform_http_client import (
    MlPlatformHttpClient,
    MlPlatformUserAuth,
    MlPlatformRequestType,
)
from mlplatform_lib.dataclass.model.model_predefinedai_dto import ModelPredefinedaiDto
from typing import List
from mlplatform_lib.utils.dataclass_utils import to_dict, from_dict


class PredefinedAIHttpClient(MlPlatformHttpClient):
    def __init__(self, mlplatform_addr):
        super().__init__(mlplatform_addr=mlplatform_addr)

    def get_models(
        self, experiment_id: int, train_id: int, auth: MlPlatformUserAuth
    ) -> List[ModelPredefinedaiDto]:
        res = self.send_request(
            "models",
            {"experiments": experiment_id, "trains": train_id},
            {},
            {},
            auth,
            MlPlatformRequestType.READ,
        )
        model_infos = []
        for model_info in res.data:
            model_infos.append(from_dict(ModelPredefinedaiDto, model_info))
        return model_infos

    def insert_model(
        self, experiment_id: int, train_id: int, model: ModelPredefinedaiDto, auth: MlPlatformUserAuth
    ) -> ModelPredefinedaiDto:
        res = self.send_request(
            "models",
            {"experiments": experiment_id, "trains": train_id},
            {},
            to_dict(model),
            auth,
            MlPlatformRequestType.CREATE,
        )
        return from_dict(ModelPredefinedaiDto, res.data)



from dataclasses import dataclass
from enum import Enum
import requests
from typing import Optional, Dict, List, Union
from pathlib import Path
from mlplatform_lib.dataclass.model import ModelInfoDto
from datetime import datetime
from mlplatform_lib.utils.dataclass_utils import to_dict, from_dict


MLPLATFORM_TIMSTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S"


class MlPlatformRequestException(Exception):
    status_code = 200

    def __init__(self, message, status_code=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code


class MlPlatformRequestType(Enum):
    CREATE = "CREATE"
    READ = "READ"
    UPDATE = "UPDATE"
    DELETE = "DELETE"


@dataclass
class MlPlatformUserAuth:
    project_id: str
    user_id: str
    authorization: str
    authorization_type: str


@dataclass
class MlPlatformRequestResult:
    data: Optional[Union[Dict, List[Dict], bytes]]
    status_code: Optional[int]


class MlPlatformHttpClient:
    def __init__(self, mlplatform_addr):
        self.base_url = mlplatform_addr

    def send_request(
        self,
        service: str,
        rest: Optional[dict],
        query: Optional[dict],
        data: Union[Optional[dict], Path],
        auth: MlPlatformUserAuth,
        request_type: MlPlatformRequestType,
    ) -> MlPlatformRequestResult:
        headers = {
            "Content-Type": "application/json",
            "Authorization": auth.authorization,
            "userId": auth.user_id,
            "projectId": auth.project_id,
            "authorizationType": auth.authorization_type,
        }

        if rest is None:
            rest = {}

        if query is None:
            query = {}

        if data is None:
            data = {}

        url = self.base_url
        for key, val in rest.items():
            url = url + "/" + str(key) + "/" + str(val)
        if service != "" and service is not None:
            url = url + "/" + service

        if len(query) != 0:
            url += "?"
            for idx, (key, val) in enumerate(query.items()):
                if idx != 0:
                    url += "&"
                url = url + str(key) + "=" + str(val)

        if request_type == MlPlatformRequestType.READ:
            response = requests.get(url, headers=headers, json=data, verify=False)
        elif request_type == MlPlatformRequestType.CREATE:
            if isinstance(data, Path):
                headers.pop("Content-Type", None)
                response = requests.post(url, headers=headers, files={"file": open(data, "rb")})
            else:
                response = requests.post(url, headers=headers, json=data, verify=False)
        elif request_type == MlPlatformRequestType.UPDATE:
            response = requests.put(url, headers=headers, json=data, verify=False)
        elif request_type == MlPlatformRequestType.DELETE:
            response = requests.delete(url, headers=headers, json=data, verify=False)
        else:
            raise MlPlatformRequestException(status_code=405, message=f"{request_type} method not allowed.")

        if response.status_code == 200:
            return MlPlatformRequestResult(response.json(), response.status_code)
        else:
            print(
                (
                    f"{service} {request_type.value} failed\n"
                    f"status_code: {response.status_code}.\n"
                    f"reason: {response.text}\n"
                )
            )
            raise MlPlatformRequestException(response.content.decode(), response.status_code)

    def upload_inference_csv(
        self, experiment_id: int, inference_id: int, inference_csv_path: str, auth: MlPlatformUserAuth
    ) -> None:
        self.send_request(
            "file",
            {"experiments": experiment_id, "inferences": inference_id},
            {},
            Path(inference_csv_path),
            auth,
            MlPlatformRequestType.CREATE,
        )

    def insert_model_info(
        self,
        experiment_id: int,
        train_id: int,
        model_id: int,
        model_info: ModelInfoDto,
        auth: MlPlatformUserAuth,
    ) -> ModelInfoDto:
        model_info.finished_time = datetime.now().strftime(MLPLATFORM_TIMSTAMP_FORMAT)

        result = self.send_request(
            "info",
            {"experiments": experiment_id, "trains": train_id, "models": model_id},
            {},
            to_dict(model_info),
            auth,
            MlPlatformRequestType.CREATE,
        )
        return from_dict(ModelInfoDto, result.data)



from mlplatform_lib.api_client import ApiClient, RunMode
from mlplatform_lib.hyperdata.hyperdata_http_client import HyperdataHttpClient, HyperdataUserAuth
from mlplatform_lib.hyperdata.hyperdata_local_checker import HyperdataLocalChecker
import os
import pathlib
import pandas as pd


class HyperdataApi:
    def __init__(self, api_client: ApiClient = None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        if self.api_client.run_mode == RunMode.LOCAL:
            self.local_checker = HyperdataLocalChecker(api_client=api_client)
            self.local_checker.run()
            self.api_client.projectId = self.local_checker.project_id
            self.api_client.userId = self.api_client.user_id
            self.api_client.Authorization = self.local_checker.access_token

        self.hyperdata_client = HyperdataHttpClient(hd_addr=self.api_client.hyperdata_addr)
        self.hyperdata_auth = HyperdataUserAuth(
            project_id=self.api_client.projectId,
            user_id=self.api_client.userId,
            authorization=self.api_client.Authorization,
        )

    def download_csv(self, do_id: int, data_rootpath: str) -> str:
        result, sep, line_delim = self.hyperdata_client.download_do_to_csv(self.hyperdata_auth, do_id)

        pathlib.Path(data_rootpath).mkdir(parents=True, exist_ok=True)
        csv_path = os.path.join(data_rootpath, "%d.csv" % do_id)

        csv_str = result.data.decode("utf-8")
        if line_delim != "\n":
            csv_str.replace(line_delim, "\n")
        with open(csv_path, "w") as f:
            f.write(csv_str)

        if sep != ",":
            data = pd.read_csv(csv_path, sep=sep)
            data.to_csv(csv_path, sep=",", index=False)

        return csv_path



from mlplatform_lib.hyperdata.hyperdata_checker import HyperdataChecker

import pytest


@pytest.mark.usefixtures("get_server_config_path")
def test_pre_check_hyperdata(get_server_config_path):
    server_config_path = get_server_config_path
    pre_check_base = HyperdataChecker(server_config_path=server_config_path)
    pre_check_base.run()



from typing import Dict
import inspect
import inflection


# dataclass field(init=False)로 지정되어 있는 변수들은
# constructor에 넣어서 생성 할 수 없다.
# 따라서, 생성자에 넘겨줄 dictionary와 setattr로 설정할 변수를 나누어 설정한다.
def from_dict(dataclass, dict: Dict):
    sig = inspect.signature(dataclass.__init__)

    init_dict = {}
    init_false_dict = {}
    for key, val in dict.items():
        # camel case to snake case
        key = inflection.underscore(key)
        if key in sig.parameters.keys():
            init_dict[key] = val
        else:
            init_false_dict[key] = val

    instance = dataclass(**init_dict)
    for key, val in init_false_dict.items():
        setattr(instance, key, val)
    return instance


def to_dict(instance):
    return {inflection.camelize(key, False): val for key, val in instance.__dict__.items()}
